import asyncio
import json
from typing import Callable
from urllib.parse import urljoin

import httpx
from bs4 import BeautifulSoup


def get_links(url: str):
    try:
        response = httpx.get(url)
        soup = BeautifulSoup(response.text, "html.parser")
        for link in soup.find_all("a"):
            href = link.get("href")
            yield urljoin(url, href)

    except httpx.HTTPError as e:
        print(f"An error occurred while fetching {url}, details: {repr(e)}")


def is_duplicate_checker() -> Callable[[str], bool]:
    url_store: dict[str, bool] = {}

    def is_duplicate(url: str) -> bool:
        is_seen = url_store.get(url, False)
        if not is_seen:
            url_store[url] = True

        return is_seen

    return is_duplicate


async def get_url_state(
    url: str, semaphore: asyncio.Semaphore, client: httpx.AsyncClient
) -> int:
    async with semaphore:
        try:
            response = await client.get(url)
            return response.status_code

        except httpx.HTTPError as err:
            print(f"error fetching {url}, details: {repr(err)}")
            return 0

        except Exception as err:
            print(f"unknown error fetching {url}, details: {repr(err)}")
            return -1


class ConsolePrinter:
    def print(self, url, result: dict[int, list]):
        print(
            f"\n################################# {url} #################################\n"
        )

        for code, urls in result.items():
            for url in urls:
                print("{:>3} | {:<40} ".format(code, url))

        print("-------------------------------------------")


class JsonPrinter:
    def __init__(self, path) -> None:
        self.file_path = path

    def print(self, url, result: dict[int, list]):

        with open(self.file_path, "w") as json_file:
            json.dump(result, json_file, indent=4)

        print(f"JSON output has been written to {self.file_path}")



import argparse
import asyncio
import collections
import os
from typing import Protocol

import httpx

from py_link_detective import __version__
from py_link_detective.detective import (
    ConsolePrinter,
    JsonPrinter,
    get_links,
    get_url_state,
    is_duplicate_checker,
)


def cli_args():
    parser = argparse.ArgumentParser(
        description="Detective: A simple link checker for websites"
    )

    # Add version argument
    parser.add_argument(
        "--version", "-v", action="version", version=f"%(prog)s v{__version__}"
    )

    # number of concurrent request
    parser.add_argument(
        "-n", type=int, default=10, help="number of concurrent request at a time"
    )

    # adding a group for printer choice(console / json)
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "-c",
        action="store_true",
        help="use console for output",
    )
    group.add_argument(
        "-j",
        action="store_true",
        help="use a json file(in the current path) for output",
    )

    # Add URL argument
    parser.add_argument(
        "url",
        metavar="url",
        type=str,
        help="The url of the website to be checked",
    )

    return parser.parse_args()


class Printer(Protocol):
    def print(self, url: str, result: dict[int, list]): ...


async def worker(args: argparse.Namespace, printer: Printer):
    is_duplicate = is_duplicate_checker()
    semaphore = asyncio.Semaphore(args.n)
    client = httpx.AsyncClient()
    results = collections.defaultdict(list)

    tasks = []
    for url in get_links(args.url):
        if is_duplicate(url):
            continue

        task = asyncio.create_task(get_url_state(url, semaphore, client))
        task.add_done_callback(lambda x, url=url: results[x.result()].append(url))  # type: ignore
        tasks.append(task)

    await asyncio.gather(*tasks)

    printer.print(args.url, results)


def main():
    args = cli_args()
    if args.j:
        printer = JsonPrinter("./output.json")
    else:
        printer = ConsolePrinter()

    asyncio.run(worker(args, printer))
